package de.hdm_stuttgart.battlearena.Model.DataStorage.Classes;

import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.ENUMs.AccountType;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.ENUMs.MapType;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Exceptions.GSONException;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Exceptions.ParserException;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Exceptions.SQLException;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Utilities.HashGenerator;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Exceptions.DatabaseException;
import de.hdm_stuttgart.battlearena.Model.DataStorage.Classes.Utilities.Parser;
import org.apache.logging.log4j.Logger;
import org.apache.logging.log4j.LogManager;

import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;

public class Persistence {

    private static final Logger log = LogManager.getLogger(Persistence.class);
    private static final Persistence persistenceSingleton = new Persistence();
    private final GsonHandler gsonHandler = new GsonHandler();
    protected OracleDB db = new OracleDB(); //evtl. Methoden von OracleDB static machen und von GSON Handler ; "protected" for testing purposes to showcase Mockito tests
    private ArrayList<MapData> coreMaps;
    private ArrayList<MapData> communityMaps;
    private PlayerStatistics statistics;
    private PlayerAccount account;
    private AppSettings settings;
    private ArrayList<MapInfo> communityMapsListRemote;  //for community map browser



    private Persistence (){}

    public static Persistence getInstance(){
        return persistenceSingleton;
    }

    public ArrayList<MapData> getCoreMaps() {
        return coreMaps;
    }

    public ArrayList<MapData> getCommunityMaps() {
        return communityMaps;
    }

    public PlayerStatistics getStatistics() {
        return statistics;
    }

    public PlayerAccount getAccount() {
        return account;
    }

    public AppSettings getSettings() {
        return settings;
    }

    public ArrayList<MapInfo> getCommunityMapsListRemote() {
        return communityMapsListRemote;
    }

    public void loadPlayerAccount() throws DatabaseException {
        try {
            account = gsonHandler.loadAccount();
            log.info("Player account successfully loaded!");
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Loading player account data from local storage failed!");
        }
    }

    public void loadPlayerStatistics() throws DatabaseException {
        try {
            if (account.getAccountType() == AccountType.LOCAL) {
                statistics = gsonHandler.loadStats();
                Parser.playerStatsValid(statistics);
            } else {
                statistics = db.getStatistics(account.getPlayerName());
                Parser.playerStatsValid(statistics);
            }
            log.info("Player statistics successfully loaded!");
        }
        catch(SQLException e){
            log.error(e);
            if(e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")){
                throw new DatabaseException("No connection to SQL server!");
            }
            else{
                throw new DatabaseException("Unknown Database Error. Retrieving statistics from SQL server failed!");
            }
        }
        catch(GSONException e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
        catch (ParserException e) {
            throw new RuntimeException("Player statistics data corrupted!");
        }
    }

    public void loadCoreMaps() throws DatabaseException {
        try {
            coreMaps = gsonHandler.loadMaps(MapType.COREMAP);
            log.info("Core-Maps successfully loaded from file!");
            //createCoreMapsList();
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Loading core maps from local storage failed!");
        }
    }

    public void loadCommunityMaps() throws DatabaseException {
        try {
            communityMaps = gsonHandler.loadMaps(MapType.COMMUNITYMAP);
            log.info("Community-Maps successfully loaded from file!");
            //createLocalCommunityMapsList();
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Loading community maps from local storage failed!");
        }
    }

    public void loadSettings() throws DatabaseException {
        try {
            settings = gsonHandler.loadSettings();
            log.info("Application settings successfully loaded from file!");
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Loading settings from local storage failed!");
        }
    }

    public void updateCoreMaps() throws DatabaseException {
        try {
            coreMaps = db.getCoreMaps();
            gsonHandler.saveMaps(coreMaps, MapType.COREMAP);
            //createCoreMapsList();
            log.info("Core-Maps successfully updated from SQL-Server!");
        }
        catch(Exception e){
            log.error(e);
            if(e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")){
                throw new DatabaseException("No connection to SQL server!");
            }
            else{
                throw new DatabaseException("Unknown Database Error. Retrieving Core-Map update failed!");
            }
        }
    }

    public void getCommunityMap(String mapID) throws DatabaseException {
        for(int i = 0; communityMaps.size() > i; i++){
            if(communityMaps.get(i).getMapID().equals(mapID)){
                throw new DatabaseException("Identical map already saved locally. See map name: " + communityMaps.get(i).getMapID());
            }
        }
        try {
            communityMaps.add(db.getCommunityMapByID(mapID));
            gsonHandler.saveMaps(communityMaps, MapType.COMMUNITYMAP);
            //createLocalCommunityMapsList();
            log.info("Community Map successfully retrieved from SQL-Server!");
        }
        catch(Exception e){
            log.error(e);
            if(e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")){
                throw new DatabaseException("No connection to SQL server!");
            }
            else if(e.getMessage().equals("No match on Database!")){
                throw new DatabaseException("No map existing for mapID: " + mapID);
            }
            else{
                throw new DatabaseException("Unknown Database Error. Retrieving Community-Map failed!");
            }
        }
    }

    public void saveSettings(){
        try{
        gsonHandler.saveSettings(settings);
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void savePlayerStatistics() throws DatabaseException{     //on shutdown of game
        try {
            Parser.playerStatsValid(statistics);
            if (account.getAccountType() == AccountType.LOCAL) {
                gsonHandler.saveStats(statistics);
            } else if (account.getAccountType() == AccountType.ONLINE) {
                db.updatePlayerStats(statistics, account);
            }
            log.info("Statistics successfully saved!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void saveCreatedMap(MapData map, boolean doUpload) throws DatabaseException {
        String part1 = "";
        try {
            for(int i = 0; communityMaps.size() > i; i++){
                if(communityMaps.get(i).getMapID().equals(map.getMapID())){
                    throw new DatabaseException("Identical map already saved locally. See map name: " + communityMaps.get(i).getMapID());
                }
            }
            communityMaps.add(map);
            gsonHandler.saveMaps(communityMaps, MapType.COMMUNITYMAP);
            log.info("Newly created map stored successfully in JSON!");
        }
        catch(Exception e){
            log.error(e);
            if(e.getMessage().contains("Identical map already saved")){
                part1 = e.getMessage();
            }
        }
        if(doUpload) {
            try {
                db.uploadCommunityMap(map);
                log.info("Newly created Community-Map successfully published!");
            }
            catch (Exception e) {
                log.error(e);
                if (e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")) {
                    throw new DatabaseException(part1 + "\nNo connection to SQL server! Saving to SQL communityMap server failed!");
                } else if (e.getMessage().contains("ORA-00001")) {
                    throw new DatabaseException(part1 + "\nMap already existing on communityMaps server!\n Look for mapID: " + map.getMapID());
                } else {
                    throw new DatabaseException(part1 + "\nUnknown Database Error. Saving to SQL communityMap server failed!");
                }
            }
        }
    }

    public void updatePlayerStatistics(int kills, int deaths, int gameTime, int blocksDestroyed, boolean gameWon) throws DatabaseException{  //after game round; only update stats in RAM, not in persistence (run savePlayerStatistics() method at end of program to save stats to persistence)
        try {
            Parser.playerStatsValid(new PlayerStatistics(0, 0, kills, deaths, blocksDestroyed, gameTime)); //temp instance of PLayerStatistics to validate new values
            statistics.addKills(kills);
            statistics.addDeaths(deaths);
            statistics.addGameTime(gameTime);
            statistics.addBlocksDestroyed(blocksDestroyed);
            if(gameWon){
                statistics.addGamesWon();
            }
            else{
                statistics.addGamesLost();
            }
            log.info("Statistics successfully updated!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void createAccount(String playerName, String password, AccountType type) throws DatabaseException {
        try {
            Parser.usernameValid(playerName);
            if(type == AccountType.LOCAL) {
                account = new PlayerAccount(playerName, "", type);
            }
            else if(type == AccountType.ONLINE) {
                Parser.passwordValid(password);
                String hashedPassword = HashGenerator.hashAndHex(password);
                db.createPlayer(playerName, hashedPassword);
                account = new PlayerAccount(playerName, hashedPassword, type);
            }
            else{
                throw new DatabaseException("Corrupted Account-Type. Please restart game.");
            }
            gsonHandler.saveAccount(account);
            log.info("Player account successfully created!");
        }
        catch(ParserException e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
        catch(SQLException e){
            log.error(e);
            if(e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")){
                throw new DatabaseException("No connection to SQL server! Do you want to create a local account?");
            }
            else if(e.getMessage().equals("ORA-00001")){
                throw new DatabaseException("Player name already taken. Please change name, use local account or login with existing account!");
            }
            else{
                throw new DatabaseException("Connection to SQL server failed! Do you want to create a local account?");
            }
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Error creating account. Please restart game or download game files again.");
        }
    }

    public void createRemoteCommunityMapsList(){        //for Map-Browser
        try {
            communityMapsListRemote = db.getCommunityMapsList();
            log.info("MapList successfully retrieved from server!");
            //log.info(communityMapsListRemote.get(0).getMapName());                 //for testing purposes
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void verifyPlayerAccount() throws DatabaseException {
        try {
            if (account.getAccountType() == AccountType.NONE) {
                throw new DatabaseException("Must create playerAccount first!");
            } else if (account.getAccountType() == AccountType.LOCAL) {
                Parser.usernameValid(account.getPlayerName());
            } else if (account.getAccountType() == AccountType.ONLINE) {
                Parser.usernameValid(account.getPlayerName());
                Parser.sha1HexHashValid(account.getAccountPassword());
                if (!(account.getAccountPassword().equals(db.checkCredentials(account.getPlayerName())))){
                    throw new DatabaseException("Locally stored password does not match online password. Please reenter credentials!");
                }
            }
            else{   //eigentlich überflüssig, da GSON schon beim Laden der Daten meckern würde, wenn er den Accounttype nicht finden kann
                throw new DatabaseException("AccountType invalid - accountData corrupted! Please create new Account or login with existing!");
            }
            log.info("Player account successfully verified!");
        }
        catch(ParserException e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
        catch(SQLException e){
            log.error(e);
            if(e.getMessage().contains("ORA-17868") | e.getMessage().contains("ORA-01017")){
                throw new DatabaseException("No connection to SQL server! Do you want to continue without tracking statistics?");
            }
            else if(e.getMessage().contains("No match on SQL database!")){
                throw new DatabaseException("No entry for player name " + account.getPlayerName() + " on SQL server. Please create account or login with different username!");
            }
            else{
                throw new DatabaseException("Connection to SQL server failed! Do you want to continue without tracking statistics");
            }
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException("Loading player account data from local storage failed! Please create new account!");
        }
    }

    public void verifyCoreMaps() throws DatabaseException{
        try {
            for (int i = 0; coreMaps.size() > i; i++) {
                Parser.mapDataValid(coreMaps.get(i).getMapData());
                Parser.mapNameValid(coreMaps.get(i).getMapName());
                Parser.sha1HexHashValid(coreMaps.get(i).getMapID());
                if(!HashGenerator.hashAndHex(coreMaps.get(i).getMapData()).equals(coreMaps.get(i).getMapID())){
                    throw new DatabaseException("mapData Hash does not match mapID!");
                }
            }
            log.info("Core-Maps data is valid!");
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
    }

    public void verifyCommunityMaps() throws DatabaseException{
        try {
            for (int i = 0; communityMaps.size() > i; i++) {
                Parser.mapDataValid(communityMaps.get(i).getMapData());
                Parser.mapNameValid(communityMaps.get(i).getMapName());
                Parser.sha1HexHashValid(communityMaps.get(i).getMapID());
                if(!HashGenerator.hashAndHex(communityMaps.get(i).getMapData()).equals(communityMaps.get(i).getMapID())){
                    throw new DatabaseException("mapData Hash does not match mapID!");
                }
            }
            log.info("Community-Maps data is valid!");
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
    }

    public void verifyAppSettings() throws DatabaseException{
        try {
            Parser.volumeValid(settings.getSfxVolume());
            Parser.volumeValid(settings.getMusicVolume());
            log.info("Settings data is valid!");
        }
        catch(Exception e){
            log.error(e);
            throw new DatabaseException(e.getMessage());
        }
    }

    public void resetPlayerAccount() throws DatabaseException{
        try{
            account = new PlayerAccount("", "", AccountType.NONE);
            gsonHandler.saveAccount(account);
            log.info("Player account successfully reset!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void resetPlayerStatistics() throws DatabaseException{
        try {
            if (account.getAccountType() == AccountType.LOCAL) {
                statistics = new PlayerStatistics(0, 0, 0, 0, 0, 0);
                gsonHandler.saveStats(statistics);
            } else if (account.getAccountType() == AccountType.ONLINE) {
                verifyPlayerAccount();
                db.resetPlayerStats(account);
                statistics = new PlayerStatistics(0, 0, 0, 0, 0, 0);
            }
            log.info("Player statistics successfully reset on local storage!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void resetSettings(){
        try {
            settings = new AppSettings(50, 50);
            gsonHandler.saveSettings(settings);
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void resetCoreMaps(){
        try {
            coreMaps = new ArrayList<MapData>();
            coreMaps.add(new MapData("09a02b54d05b5b7ebc29a4383ca12d3dda846b72", "Arena1", 18, 18, "12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12"));
            coreMaps.add(new MapData("0ab15557ab6dc4be60dfe6a9b0288bac3036bd97", "Arena2", 18, 18, "12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12 12 1 1 1 1 1 1 1 5 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12"));
            gsonHandler.saveMaps(coreMaps, MapType.COREMAP);
            log.info("Core maps successfully reset on local storage!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void resetCommunityMaps(){
        try {
            communityMaps = new ArrayList<MapData>();
            communityMaps.add(new MapData("fbf44184867512faecc195ae75ca55d5ab7bad2d", "Arena3", 18, 18, "12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12 12 1 1 1 1 1 1 1 8 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12"));
            communityMaps.add(new MapData("1c23b362fd666c5fb7ed60ca611b17246424e49f", "Arena4", 18, 18, "12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12 12 1 1 1 1 1 1 1 8 4 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 1 12 12 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 10 12"));
            gsonHandler.saveMaps(communityMaps, MapType.COMMUNITYMAP);
            log.info("Community maps successfully reset on local storage!");
        }
        catch(Exception e){
            log.error(e);
        }
    }

    public void login(String userName, String password) throws NoSuchAlgorithmException, RuntimeException, DatabaseException {
        try {
            String hash = HashGenerator.hashAndHex(password);
            if (hash.equals(db.checkCredentials(userName))) {
                account = new PlayerAccount(userName, hash, AccountType.ONLINE);
                gsonHandler.saveAccount(account);
                log.info("Login successful!");
            } else {
                throw new DatabaseException("Wrong password!");
            }
        }
        catch (GSONException | SQLException e) {
            throw new RuntimeException(e);
        }
    }

}
